import json


ROOT_HEAD = 0
NULL = "#NULL"

ID = "id"
WORD = "word"
LEMMA = "lemma"
UPOS = "upos"
XPOS = "xpos"
FEATS = "feats"
HEAD = "head"
DEPREL = "deprel"
DEPS = "deps"
MISC = "misc"
DEEPSLOT = "deepslot"
SEMCLASS = "semclass"

MANDATORY_TAGS = [ID, WORD]
OPTIONAL_TAGS = [LEMMA, UPOS, XPOS, FEATS, HEAD, DEPREL, DEPS, MISC, DEEPSLOT, SEMCLASS]

# ========================= Field level =========================

def parse_nullable(field: str) -> str | None:
    if field in ["", "_"]:
        return None
    return field


def is_null_index(x: str) -> bool:
    try:
        a, b = x.split('.')
        return a.isdecimal() and b.isdecimal()
    except ValueError:
        return False

def is_range_index(x: str) -> bool:
    try:
        a, b = x.split('-')
        return a.isdecimal() and b.isdecimal()
    except ValueError:
        return False

def parse_id(field: str) -> str:
    if not field.isdecimal() and not is_null_index(field) and not is_range_index(field):
        raise ValueError(f"Incorrect token id: {field}.")
    return field


def parse_word(field: str) -> str:
    if field == "":
        raise ValueError(f"Token form cannot be empty.")
    return field


def parse_joint_field(field: str, inner_sep: str, outer_sep: str) -> dict[str, str]:
    """
    Parse joint field with two kinds of separators (inner and outer) and return a dict.

    E.g.
    >>> parse_joint_field("Mood=Ind|Number=Sing|Person=3", '=', '|')
    {'Mood': 'Ind', 'Number': 'Sing', 'Person': '3'}
    >>> parse_joint_field("26:conj|18.1:advcl:while", ':', '|')
    {'26': 'conj', '18.1': 'advcl:while'}
    """
    if inner_sep not in field:
        raise ValueError(f"Non-empty field {field} must contain {inner_sep} separator.")

    field_dict = {}
    for item in field.split(outer_sep):
        key, value = item.split(inner_sep, 1)
        if key in field_dict:
            raise ValueError(f"Field {field} has duplicate key {key}.")
        field_dict[key] = value
    return field_dict


def parse_feats(field: str) -> str | None:
    if field in ["", "_"]:
        return None

    # validate feats are parsable
    parse_joint_field(field, inner_sep='=', outer_sep='|')
    
    return field


def parse_head(field: str) -> int | None:
    if field in ["", "_"]:
        return None

    if not field.isdecimal():
        raise ValueError(f"Non-empty head must be a decimal number, not {field}.")

    head = int(field)
    if head < 0:
        raise ValueError(f"Non-empty head must be a positive integer, not {head}.")

    return head


def parse_deps(field: str) -> str | None:
    if field in ["", "_"]:
        return None

    token_deps = parse_joint_field(field, inner_sep=':', outer_sep='|')

    # Validate deps.
    if len(token_deps) == 0:
        raise ValueError(f"Empty deps are not allowed: {field}.")
    for head in token_deps:
        if not head.isdecimal() and not is_null_index(head):
            raise ValueError(f"Deps head must be either integer or float (x.1), not {head}.")
    return json.dumps(token_deps)

# ========================= Token level =========================

Token = dict[str, str]

def validate_null_token(token: Token):
    if token[WORD] != NULL:
        raise ValueError(f"Null token must have #NULL form, not {token[WORD]}.")
    if HEAD in token and token[HEAD] is not None:
        raise ValueError(f"Null token must have no head, but found {token[HEAD]}.")
    if DEPREL in token and token[DEPREL] is not None:
        raise ValueError(f"Null token must have no deprel, but found {token[DEPREL]}.")
    if MISC in token and token[MISC] != 'ellipsis':
        raise ValueError(f"Null token must have 'ellipsis' misc, not {token[MISC]}.")

def validate_range_token(token: Token):
    if LEMMA in token and token[LEMMA] != '_':
        raise ValueError(f"Range token lemma must be _, but found {token[LEMMA]}.")
    if UPOS in token and token[UPOS] is not None:
        raise ValueError(f"Range token upos must be _, but found {token[UPOS]}.")
    if XPOS in token and token[XPOS] is not None:
        raise ValueError(f"Range token xpos must be _, but found {token[XPOS]}.")
    if FEATS in token and token[FEATS] is not None:
        raise ValueError(f"Range token feats must be _, but found {token[FEATS]}.")
    if HEAD in token and token[HEAD] is not None:
        raise ValueError(f"Range token head must be _, but found {token[HEAD]}.")
    if DEPREL in token and token[DEPREL] is not None:
        raise ValueError(f"Range token deprel must be _, but found {token[DEPREL]}.")
    if MISC in token and token[MISC] is not None:
        raise ValueError(f"Range token misc must be _, but found {token[MISC]}.")
    if DEEPSLOT in token and token[DEEPSLOT] is not None:
        raise ValueError(f"Range token deepslot must be _, but found {token[DEEPSLOT]}.")
    if SEMCLASS in token and token[SEMCLASS] is not None:
        raise ValueError(f"Range token semclass must be _, but found {token[SEMCLASS]}.")

def validate_regular_token(token: Token):
    if HEAD in token and token[HEAD] == token[ID]:
        raise ValueError(f"Self-loop detected in head.")
    if DEPS in token and token[DEPS] is not None:
        for head in json.loads(token[DEPS]):
            if head == token[ID]:
                raise ValueError(f"Self-loop detected in deps. head: {head}, id: {token[ID]}")

def validate_token(token: Token):
    if token[ID].isdecimal():
        validate_regular_token(token)
    elif is_range_index(token[ID]):
        validate_range_token(token)
    elif is_null_index(token[ID]):
        validate_null_token(token)
    else:
        raise ValueError(f"Incorrect token id: {token[ID]}.")


def parse_token(line: str, tags_to_parse: list[str], validate: bool) -> Token:
    try:
        fields = [field.strip() for field in line.split("\t")]
        if len(fields) < len(tags_to_parse):
            raise SyntaxError(
                f"line {line} must have at leasts {len(tags_to_parse)} fields, not {len(fields)}"
            )

        token = {}
        if ID in tags_to_parse:
            token[ID] = parse_id(fields[0])
        if WORD in tags_to_parse:
            token[WORD] = parse_word(fields[1])
        if LEMMA in tags_to_parse:
            token[LEMMA] = fields[2] if fields[2] != "" else None
        if UPOS in tags_to_parse:
            token[UPOS] = parse_nullable(fields[3])
        if XPOS in tags_to_parse:
            token[XPOS] = parse_nullable(fields[4])
        if FEATS in tags_to_parse:
            token[FEATS] = parse_feats(fields[5])
        if HEAD in tags_to_parse:
            token[HEAD] = parse_head(fields[6])
        if DEPREL in tags_to_parse:
            token[DEPREL] = parse_nullable(fields[7])
        if DEPS in tags_to_parse:
            token[DEPS] = parse_deps(fields[8])
        if MISC in tags_to_parse:
            token[MISC] = parse_nullable(fields[9])
        if DEEPSLOT in tags_to_parse:
            token[DEEPSLOT] = parse_nullable(fields[10])
        if SEMCLASS in tags_to_parse:
            token[SEMCLASS] = parse_nullable(fields[11])

        if validate:
            validate_token(token)

    except Exception as e:
        raise RuntimeError(f"Validation failed on token {fields[0]}:\n{e}")

    return token

# ========================= Sentence level =========================

Sentence = dict[str, list[str]]

def validate_sentence(sentence: Sentence):
    # Ensure the sentence is not empty
    num_words = len(sentence[WORD])
    if num_words == 0:
        raise ValueError("Empty sentence.")

    for tag in OPTIONAL_TAGS:
        if tag in sentence and len(sentence[tag]) != num_words:
            raise RuntimeError(f"Field {tag} has inconsistent length.")

    # Validate sentence heads and ids agreement.
    ids = set(sentence[ID])
    int_ids = {int(idtag) for idtag in sentence[ID] if idtag.isdecimal()}

    contiguous_int_ids = set(range(min(int_ids), max(int_ids) + 1))
    if int_ids != contiguous_int_ids:
        raise RuntimeError(f"ids are not contiguous: {sorted(int_ids)}.")

    if HEAD in sentence:
        # Check 1
        has_labels = any(head is not None for head in sentence[HEAD])
        roots_count = sum(head == ROOT_HEAD for head in sentence[HEAD])
        if has_labels and roots_count != 1:
            raise RuntimeError(
                f"There must be exactly one ROOT in a sentence, but found {roots_count}."
            )
        # Check 2
        sentence_heads = set(
            head
            for head in sentence[HEAD]
            if head is not None and head != ROOT_HEAD
        )
        excess_heads = sentence_heads - int_ids
        if excess_heads:
            raise RuntimeError(
                f"Heads are inconsistent with sentence ids. "
                f"Excessive heads: {excess_heads}."
            )

    if DEPS in sentence:
        sentence_deps_heads = {
            head
            for deps in sentence[DEPS] if deps is not None
            for head in json.loads(deps) if head != str(ROOT_HEAD)
        }
        excess_deps_heads = sentence_deps_heads - ids
        if excess_deps_heads:
            raise RuntimeError(
                f"Deps heads are inconsistent with sentence ids. "
                f"Excessive heads: {excess_deps_heads}."
            )


def parse_sentence(
    token_lines: list[str],
    metadata: dict,
    tags_to_parse: list[str],
    validate: bool
) -> Sentence:
    try:
        tokens = [
            parse_token(token_line, tags_to_parse, validate)
            for token_line in token_lines
        ]
        if validate:
            sentence = {
                tag: [token[tag] for token in tokens]
                for tag in tags_to_parse
            }
            validate_sentence(sentence)

    except Exception as e:
        raise RuntimeError(f"Validation failed on sentence {metadata.get('sent_id', None)}:\n{e}")

    return metadata | {"tokens": tokens}


def parse_incr(filepath: str, optional_tags_to_parse: list[str], validate: bool):
    """
    Generator that parses a CoNLL-U Plus file in CoBaLD format and yields one sentence at a time.
    
    Each sentence is represented as a dictionary:
    {
        "sent_id": <sent_id from metadata or None>,
        "text": <text from metadata or None>,
        ID: list of sentence tokens ids,
        WORD: list of tokens forms,
        ...
        <optional tags to parse, specified in `optional_tags_to_parse` argument>
        ...
    }

    The input file must have metadata lines starting with '#' (e.g., "# sent_id = 1")
    and token lines. Sentence blocks must be separated by blank lines.
    """
    metadata = {}
    token_lines = []

    for tag in optional_tags_to_parse:
        if tag not in OPTIONAL_TAGS:
            raise ValueError(f"Invalid tags: {tag}, must be one of {OPTIONAL_TAGS}")

    tags_to_parse = MANDATORY_TAGS + optional_tags_to_parse

    with open(filepath, "r", encoding="utf-8") as file:
        for line in file:
            line = line.strip()
            if not line:
                # End of a sentence block.
                if token_lines:
                    yield parse_sentence(token_lines, metadata, tags_to_parse, validate)
                    metadata = {}
                    token_lines = []
                continue
            if line.startswith("#"):
                # Process metadata lines.
                if "=" in line:
                    # Remove the '#' and split by the first '='.
                    key, value = line[1:].split("=", 1)
                    key = key.strip()
                    value = value.strip()
                    if key in ("sent_id", "text"):
                        metadata[key] = value
                continue
            # Accumulate token lines.
            token_lines.append(line)

    # Yield any remaining sentence at the end of the file.
    if token_lines:
        yield parse_sentence(token_lines, metadata, tags_to_parse, validate)


import sys
import json
from collections import defaultdict

from tqdm import tqdm
import numpy as np
from sklearn.metrics import f1_score

from .taxonomy import Taxonomy


NULL = "#NULL"


def is_range_id(x: str) -> bool:
    try:
        a, b = x.split('-')
        return a.isdecimal() and b.isdecimal()
    except ValueError:
        return False

def parse_feats(feats: str) -> dict:
    field_dict = {}
    for item in feats.split('|'):
        key, value = item.split('=')
        field_dict[key] = value
    return field_dict

def jaccard_score(a: set, b: set, aggregate_fn: callable = len) -> float:
    return aggregate_fn(a & b) / aggregate_fn(a | b)


class CobaldScorer:
    def __init__(
        self,
        taxonomy: Taxonomy = None,
        semclasses_out_of_taxonomy: set = {},
        lemma_weights: dict[str, float] = None,
        feats_weights: dict[str, float] = None
    ):
        self.taxonomy = taxonomy
        self.semclasses_out_of_taxonomy = set(semclasses_out_of_taxonomy)
        self.lemma_weights = lemma_weights
        self.feats_weights = feats_weights

    def score_lemma(self, test: str, gold: str) -> float:
        if test["lemma"] is None or gold["lemma"] is None:
            return float(test["lemma"] == gold["lemma"])

        normalize = lambda word: word.lower().replace('ё', 'е')
        score = normalize(test["lemma"]) == normalize(gold["lemma"])

        # 'lemma' scores are weighted, because we want immutable parts of speech
        # to affect lemmatization score less than mutable ones (which are, obviously,
        # harder to lemmatize).
        if self.lemma_weights and gold["upos"]:
            score *= self.lemma_weights[gold["upos"]]
        return score

    def score_upos(self, test: dict, gold: dict) -> float:
        return float(test["upos"] == gold["upos"])

    def score_xpos(self, test: dict, gold: dict) -> float:
        return float(test["xpos"] == gold["xpos"])

    def score_feats(self, test: dict, gold: dict) -> float:
        if not test["feats"] or not gold["feats"]:
            return test["feats"] == gold["feats"]

        assert isinstance(test["feats"], str) and isinstance(gold["feats"], str)
        test_feats = parse_feats(test["feats"])
        gold_feats = parse_feats(gold["feats"])

        if len(test_feats) == 0 or len(gold_feats) == 0:
            return float(test_feats == gold_feats)

        test_feats_items = set(test_feats.items())
        gold_feats_items = set(gold_feats.items())

        sum_weights = lambda feats: sum(self.feats_weights[cat] for cat, _ in feats)
        return jaccard_score(
            test_feats_items,
            gold_feats_items,
            sum_weights if self.feats_weights else len
        )

    def score_head(self, test: dict, gold: dict) -> float:
        return float(test["head"] == gold["head"])

    def score_deprel(self, test: dict, gold: dict) -> float:
        return float(test["head"] == gold["head"] and test["deprel"] == gold["deprel"])

    def score_deps(self, test: dict, gold: dict, use_labels: bool) -> float:
        if not test["deps"] or not gold["deps"]:
            return test["deps"] == gold["deps"]

        assert isinstance(test["deps"], str) and isinstance(gold["deps"], str)
        test_deps = json.loads(test["deps"])
        gold_deps = json.loads(gold["deps"])

        if len(test_deps) == 0 or len(gold_deps) == 0:
            return float(test_deps == gold_deps)

        test_arcs = set(test_deps.items() if use_labels else test_deps.keys())
        gold_arcs = set(gold_deps.items() if use_labels else gold_deps.keys())
        return jaccard_score(test_arcs, gold_arcs)

    def score_misc(self, test: str, gold: str) -> float:
        return float(test["misc"] == gold["misc"])

    def score_deepslot(self, test: dict, gold: dict) -> float:
        return float(test["deepslot"] == gold["deepslot"])

    def score_semclass(self, test: dict, gold: dict) -> float:
        # If gold semclass is out of taxomony, simply compare strings
        if gold["semclass"] in self.semclasses_out_of_taxonomy:
            return float(test["semclass"] == gold["semclass"])

        if not self.taxonomy:
            raise RuntimeError("Taxonomy is missing")
        if not self.taxonomy.has_semclass(gold["semclass"]):
            raise ValueError(f"Unknown gold semclass encountered: {gold["semclass"]}")
        if not self.taxonomy.has_semclass(test["semclass"]):
            return 0.

        semclasses_distance = self.taxonomy.calc_path_length(test["semclass"], gold["semclass"])
        # If distance is 0 then test_semclass == gold_semclass, so score is 1.
        # If they are different, the penalty is proportional to their distance.
        # If they are in different trees, then distance is inf, so score is 0.
        return 1 / (1 + semclasses_distance)

    def score_sentences(self, test_sentences, gold_sentences) -> tuple[float]:
        # name, column, scoring function
        scoring_info = [
            ("lemma weight. accuracy", "lemma", self.score_lemma),
            ("upos accuracy", "upos", self.score_upos),
            ("xpos accuracy", "xpos", self.score_xpos),
            ("feats weight. jaccard", "feats", self.score_feats),
            ("UAS", "head", self.score_head),
            ("LAS", "deprel", self.score_deprel),
            ("EUAS", "deps", lambda t, g: self.score_deps(t, g, use_labels=False)),
            ("ELAS", "deps", lambda t, g: self.score_deps(t, g, use_labels=True)),
            ("misc accuracy", "misc", self.score_misc),
            ("deepslot accuracy", "deepslot", self.score_deepslot),
            ("semclass similarity", "semclass", self.score_semclass)
        ]
        # Grammatical scores.
        # Use lists and 'np.sum' instead of float summation (+=) for numerical stability.
        test_scores = defaultdict(list)
        # Some scores are weighted and can be greater/less than 1, so if we want final
        # averaged scores to be in [0..1] range, we have to scale them by maximum possible
        # (gold) scores.
        gold_scores = defaultdict(list)

        test_nulls, gold_nulls = [], []

        for test_sentence, gold_sentence in tqdm(
            zip(test_sentences, gold_sentences, strict=True), file=sys.stdout
        ):
            # Filter out range tokens.
            test_tokens = [
                token for token in test_sentence["tokens"]
                if not is_range_id(token["id"])
            ]
            gold_tokens = [
                token for token in gold_sentence["tokens"]
                if not is_range_id(token["id"])
            ]
            # Test and gold sentence may have different lengths due to null tokens, so align them.
            try:
                test_tokens_aligned, gold_tokens_aligned = self._align_sentences(
                    test_tokens, gold_tokens
                )
            except Exception as e:
                raise RuntimeError(f"Sentence {gold_sentence.get('sent_id', '?')}:\n{e}")

            for test_token, gold_token in zip(
                test_tokens_aligned, gold_tokens_aligned, strict=True
            ):
                for score_name, column, score_fn in scoring_info:
                    # Allow partial evaluation (i.e. evaluation on subset of columns)
                    if column not in set(test_token) | set(gold_token):
                        continue

                    # Token mismatch
                    if test_token is None or gold_token is None:
                        test_score = 0
                        gold_score = 1
                    else:
                        test_score = score_fn(test_token, gold_token)
                        gold_score = score_fn(gold_token, gold_token)
                        assert 0. <= test_score <= 1.0
                        assert 0. <= gold_score <= 1.0
                    test_scores[score_name].append(test_score)
                    gold_scores[score_name].append(gold_score)

                    # Track nulls separately
                    test_nulls.append(test_token["word"] == NULL if test_token else 0)
                    gold_nulls.append(gold_token["word"] == NULL if gold_token else 0)

        average_scores = {
            name: np.mean(test_scores[name]) / np.mean(gold_scores[name])
            for name in test_scores.keys()
        }
        average_scores["null f1"] = f1_score(gold_nulls, test_nulls, zero_division=1.0)
        average_scores["total"] = np.mean([score for score in average_scores.values()])

        for score in average_scores.values():
            assert 0. <= score <= 1.

        return average_scores

    @staticmethod
    def _align_sentences(lhs: list[dict], rhs: list[dict]) -> tuple[list]:
        """
        Aligns two sequences of tokens. None token is inserted where needed.
        Example:
        >>> test_words = ["How", "did", "this", "#NULL", "happen"]
        >>> gold_words = ["How", "#NULL", "did", "this", "happen"]
        >>> _align_sentences(test_words, gold_words)
        (["How",    None, "did", "this", "#NULL", "happen"],
         ["How", "#NULL", "did", "this",    None, "happen"])
        """
        lhs_aligned, rhs_aligned = [], []

        i, j = 0, 0

        i_exhausted, j_exhausted = False, False
        while not i_exhausted and not j_exhausted:
            if lhs[i]["word"] == rhs[j]["word"]:
                lhs_aligned.append(lhs[i])
                rhs_aligned.append(rhs[j])
                i += 1
                j += 1
            elif lhs[i]["word"] == NULL and rhs[j]["word"] != NULL:
                lhs_aligned.append(lhs[i])
                rhs_aligned.append(None)
                i += 1
            elif lhs[i]["word"] != NULL and rhs[j]["word"] == NULL:
                lhs_aligned.append(None)
                rhs_aligned.append(rhs[j])
                j += 1
            else:
                raise RuntimeError(
                    f"Test-gold words mismatch: {lhs[i]['word']} != {rhs[j]['word']}"
                )
            if len(lhs) <= i:
                i_exhausted = True
                i = len(lhs) - 1
            if len(rhs) <= j:
                j_exhausted = True
                j = len(rhs) - 1

        assert len(lhs_aligned) == len(rhs_aligned)
        return lhs_aligned, rhs_aligned


import pandas as pd
from collections import Counter

from .lca import find_lca


# Common functions.
def is_unique(series: pd.Series) -> bool:
    return len(series.unique()) == len(series)

def enumerate_column(df: pd.DataFrame, column_name: str) -> dict[int, any]:
    return {index: value for index, value in enumerate(df[column_name])}

def enumerate_column_inv(df: pd.DataFrame, column_name: str) -> dict[int, any]:
    # Ensure column has unique values.
    assert is_unique(df[column_name])
    return {value: index for index, value in enumerate(df[column_name])}


class Taxonomy:
    """
    Taxonomy of semantic classes.
    """
    SEMCLASS_TYPE_ID = 0

    def __init__(self, taxonomy_file: str):
        taxonomy_df = self.load(taxonomy_file)
        self.parents = self.extract_parents(taxonomy_df)
        self.depths = self.calc_depths(self.parents)
        # Semclass tricks.
        semclass_counter = Counter(taxonomy_df["Name"])
        for i, (semclass, count) in enumerate(semclass_counter.items()):
            # All taxonomy semclasses must be unique.
            assert count == 1
        self.semclass_to_idx = {value: index for index, value in enumerate(taxonomy_df["Name"])}

    @staticmethod
    def load(taxonomy_file_csv: str) -> pd.DataFrame:
        print(f"taxonomy_file_csv: {taxonomy_file_csv}")
        taxonomy = pd.read_csv(
            taxonomy_file_csv,
            dtype={
                'ID': int,
                'ParentID': 'Int64',
                'Depth': int,
                'Name': str,
            }
        )
        # Ignore Depth because it might be broken.
        return taxonomy[['ID', 'ParentID', 'Name']]

    @staticmethod
    def extract_parents(taxonomy: pd.DataFrame) -> list[int]:
        # Enumerate nodes in a contiguous manner from 0 to max(taxonomy.index)
        node_id_to_idx = enumerate_column_inv(taxonomy, "ID")
        node_id_to_parent_id = taxonomy.set_index("ID")['ParentID'].to_dict()
        node_idx_to_parent_idx = {
            node_id_to_idx[node_id]: (node_id_to_idx[parent_id] if not pd.isna(parent_id) else -1)
            for node_id, parent_id in node_id_to_parent_id.items()
        }

        # Continuous array of parents, i.e. parents[i] = {parent of node with index i}.
        parents = [0] * len(taxonomy)
        for node_idx in range(len(taxonomy)):
            parents[node_idx] = node_idx_to_parent_idx[node_idx]

        return parents

    @staticmethod
    def calc_depths(parents: list[int]) -> list[int]:
        num_of_nodes = len(parents)
        # Array of depths, i.e. depths[i] = {depth of node with index i}.
        depths = [-1] * num_of_nodes

        # Initialize depth for roots.
        for i in range(num_of_nodes):
            if parents[i] == -1:
                depths[i] = 0

        # "Relaxation" algorithm. Ineffective!
        # Every iteration of 'while' loop the next level of nodes has depths assigned.
        # I.e. after the first iteration 'depths' array has only 0s and 1s,
        # after the second iteration 'depths' array has all 0s, 1s and 2s correct, etc.
        has_unassigned_depth = True
        while has_unassigned_depth:
            has_unassigned_depth = False
            for i in range(num_of_nodes):
                if depths[i] == -1:
                    has_unassigned_depth = True
                    parent_depth = depths[parents[i]]
                    if parent_depth != -1:
                        depths[i] = parent_depth + 1

        # Check there are no uninitialized depths.
        for i in range(num_of_nodes):
            assert 0 <= depths[i]

        return depths

    def has_semclass(self, semclass: str) -> bool:
        return semclass in self.semclass_to_idx

    def calc_path_length(self, semclass1: str, semclass2: str) -> int:
        """
        Return length of shortest (since taxonomy is a set of trees, it's also unique) path
        between two semantic classes in taxonomy.
        If classes are in different trees, return infinity.
        """
        semclass1_idx = self.semclass_to_idx[semclass1]
        semclass2_idx = self.semclass_to_idx[semclass2]

        # Path between u and v in a tree = path from u to LCA(u,v) and path from LCA(u, v) to v.
        # So find LCA(u, v) first.
        lca_index = find_lca(semclass1_idx, semclass2_idx, self.parents, self.depths)

        if lca_index is None:
            # Classes are in different trees.
            return float("inf")

        lca_depth = self.depths[lca_index]
        semclass1_depth = self.depths[semclass1_idx]
        semclass2_depth = self.depths[semclass2_idx]

        # Since lca is an ancestor for both semclass1 and semclass2, it should be higher in taxonomy.
        assert lca_depth <= semclass1_depth and lca_depth <= semclass2_depth

        semclass1_rel_depth = semclass1_depth - lca_depth
        semclass2_rel_depth = semclass2_depth - lca_depth

        return semclass1_rel_depth + semclass2_rel_depth


def find_lca(u: int, v: int, parents: list[int], depths: list[int]) -> int:
    """
    Find Lowest Common Ancestor (LCA) in a forest formed by child-parent relations.
    """
    def uplift(node_idx: int, height: int):
        nonlocal parents, depths

        assert height >= 0

        for _ in range(height):
            node_idx = parents[node_idx]

        return node_idx

    # Make depths equal.
    if depths[u] < depths[v]:
        v = uplift(v, depths[v] - depths[u])
    elif depths[u] > depths[v]:
        u = uplift(u, depths[u] - depths[v])

    assert depths[u] == depths[v], f"After uplift: {depths[u]} and {depths[v]}"
    current_depth = depths[u]

    while u != v:
        # Hit the root -> u and v are in different trees -> no LCA.
        if current_depth == 0:
            u = v = None
            break

        # Uplift vertexes simultaneously. 
        u = parents[u]
        v = parents[v]
        current_depth -= 1

    assert u == v
    lca = u
    return lca


import os
import argparse
import json

from .parsing import parse_incr, OPTIONAL_TAGS
from .scorer import CobaldScorer
from .taxonomy import Taxonomy


def load_dict_from_json(json_filepath: str) -> dict:
    with open(json_filepath, "r") as file:
        data = json.load(file)
    return data


def main():
    parser = argparse.ArgumentParser(description="CoBaLD evaluation script")

    parser.add_argument(
        'test_file',
        type=str,
        help='Test file in CoBaLD format with predicted tags.'
    )
    parser.add_argument(
        'gold_file',
        type=str,
        help="Gold file in CoBaLD format with true tags."
    )
    parser.add_argument(
        'output_file',
        type=str,
        help='Output JSON file where scoring results will be written to.'
    )
    script_dir = os.path.dirname(__file__)
    parser.add_argument(
        '--taxonomy_file',
        type=str,
        help="File in CSV format with semantic class taxonomy.",
        default=os.path.join(script_dir, "res", "hyperonims_hierarchy.csv")
    )
    parser.add_argument(
        '--lemma_weights_file',
        type=str,
        help="JSON file with 'POS' -> 'lemma weight for this POS' relations.",
        default=os.path.join(script_dir, "res", "lemma_weights.json")
    )
    parser.add_argument(
        '--feats_weights_file',
        type=str,
        help="JSON file with 'grammatical category' -> 'weight of this category' relations.",
        default=os.path.join(script_dir, "res", "feats_weights.json")
    )
    parser.add_argument(
        "--tags",
        nargs="?",
        type=str,
        default=OPTIONAL_TAGS,
        choices=OPTIONAL_TAGS,
        help=(
            "Tags to include in dataset, e.g. `heads deprels deps`."
            "By default, all CoBaLD tags are used."
        )
    )
    args = parser.parse_args()

    print(f"Loading taxonomy from {args.taxonomy_file}")
    semclass_taxonomy = Taxonomy(args.taxonomy_file)

    print(f"Loading lemma weights from {args.lemma_weights_file}")
    lemma_weights = load_dict_from_json(args.lemma_weights_file)

    print(f"Loading feats weights from {args.feats_weights_file}")
    feats_weights = load_dict_from_json(args.feats_weights_file)

    scorer = CobaldScorer(
        semclass_taxonomy,
        semclasses_out_of_taxonomy={None},
        lemma_weights=lemma_weights,
        feats_weights=feats_weights
    )

    print("Evaluating")
    test_sentences = parse_incr(
        args.test_file,
        args.tags or [],
        validate=False
    )
    gold_sentences = parse_incr(
        args.gold_file,
        args.tags or [],
        validate=False
    )
    scores = scorer.score_sentences(test_sentences, gold_sentences)

    with open(args.output_file, 'w') as f:
        json.dump(scores, f)

    print(f"Done, results written to {args.output_file}")


if __name__ == "__main__":
    main()